#ifdef EXP_WIN32

#include "NativeRegistry.h"
#include <windows.h>


typedef BOOL SubKeyCallback(PCWSTR name, PVOID usrPtr);
typedef BOOL ValueCallback(DWORD type, PCWSTR name, PBYTE data,
    DWORD dataLen, PVOID usrPtr);


typedef struct
{
    JNIEnv* e;
    jobject o;
}
JNIData;


static HKEY decodeHKId(jbyte hkid)
{
    switch(hkid)
    {
        case com_parasoft_xtest_common_nativecode_NativeRegistry_HKCR:
        	return HKEY_CLASSES_ROOT;
        	
        case com_parasoft_xtest_common_nativecode_NativeRegistry_HKCU:
        	return HKEY_CURRENT_USER;
        	
        case com_parasoft_xtest_common_nativecode_NativeRegistry_HKLM:
        	return HKEY_LOCAL_MACHINE;
        	
        case com_parasoft_xtest_common_nativecode_NativeRegistry_HKU:
        	return HKEY_USERS;
    };
    
    return INVALID_HANDLE_VALUE;
}


static jbyte encodeValueType(DWORD type)
{
    switch(type)
    {
        case REG_BINARY:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_BINARY;
        	
        case REG_DWORD:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_DWORD;
        	
        case REG_DWORD_BIG_ENDIAN:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_DWORD_BIG_ENDIAN;
        	
        case REG_SZ:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_SZ;
        	
        case REG_MULTI_SZ:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_MULTI_SZ;
        	
        case REG_EXPAND_SZ:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_EXPAND_SZ;
        	
        case REG_LINK:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_LINK;
        	
        case REG_RESOURCE_LIST:
        	return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_RESOURCE_LIST;
    };
    
    return com_parasoft_xtest_common_nativecode_NativeRegistry_REG_NONE;
}     


static BYTE queryValue0(HKEY key, PCWSTR name, PDWORD typePtr,
    PBYTE* dataPtr, PDWORD dataLenPtr)
{
    if (RegQueryValueExW(key, name, NULL, typePtr, NULL, dataLenPtr)
            != NO_ERROR)
    {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_CANT_QUERY_VALUE;
    }

    *dataPtr = malloc(*dataLenPtr);
    
    RegQueryValueExW(key, name, NULL, NULL, *dataPtr, dataLenPtr);

    return com_parasoft_xtest_common_nativecode_NativeRegistry_SUCCESS;
}


static BYTE enumSubKeys0(HKEY key, SubKeyCallback callback, PVOID usrPtr)
{
    BYTE r = com_parasoft_xtest_common_nativecode_NativeRegistry_SUCCESS;
    UINT i = 0;
    
    while(TRUE) {
        WCHAR name[MAX_PATH + 1];

        const LONG result = RegEnumKeyW(key, i, name, MAX_PATH + 1);

        if (result == NO_ERROR) {
            if (!callback(name, usrPtr)) {
                break;
            }
        } else if (result == ERROR_NO_MORE_ITEMS) {
            break;
        } else {
            r = com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_DURING_ENUM;
        }
        
        ++i;
    }
    
    return r;
}


static BYTE enumValues0(HKEY key, ValueCallback callback, PVOID usrPtr)
{
    BOOL r = com_parasoft_xtest_common_nativecode_NativeRegistry_SUCCESS;
    LONG result;
    UINT i;
    DWORD no;
    DWORD maxNameLen;
    DWORD maxDataLen;
    DWORD maxNameLen0;
    DWORD maxDataLen0;
    PWSTR name;
    PBYTE data;

    result = RegQueryInfoKey(key, NULL, NULL, NULL, NULL, NULL, NULL,
        &no, &maxNameLen, &maxDataLen, NULL, NULL);

    if (result != NO_ERROR) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_CANT_QUERY_KEY_INFO;
    }
    
    maxNameLen0 = maxNameLen + 1;
    maxDataLen0 = maxDataLen + 1;
    
    name = malloc(maxNameLen0);
    data = malloc(maxDataLen0);
    
    for (i = 0; i < no; ++i) {
        DWORD nameLen = maxNameLen0;
        DWORD dataLen = maxDataLen0;
        DWORD type;

        result = RegEnumValueW(key, i, name, &nameLen, NULL, &type,
            data, &dataLen);

        if (result == NO_ERROR) {
            if (!callback(type, name, data, dataLen, usrPtr)) {
                break;
            }
        } else {
            r = com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_DURING_ENUM;
        }
    }
    
    free(name);
    free(data);
    
    return (BYTE)r;
}


static HKEY openKey(HKEY root, PCWSTR path, BOOL* notRootPtr)
{
    HKEY key;
    
    *notRootPtr = (BOOL) lstrlenW(path);
    
    if (*notRootPtr) {
        const LONG result = RegOpenKeyW(root, path, &key);
        if (result != NO_ERROR) {
            key = INVALID_HANDLE_VALUE;
        }
    } else {
        key = root;
    }

    return key;
}


static BYTE queryValue(HKEY root, PCWSTR path, PCWSTR name, PDWORD typePtr,
    PBYTE* dataPtr, PDWORD dataLenPtr)
{
    BYTE r;
    BOOL notRoot;
    const HKEY key = openKey(root, path, &notRoot);

    if (key == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_CANT_OPEN_KEY;
    }
 
    r = queryValue0(key, name, typePtr, dataPtr, dataLenPtr);
    
    if (notRoot) {
        RegCloseKey(key);
    }
    
    return r;
}


static BYTE enumSubKeys(HKEY root, PCWSTR path,
    SubKeyCallback callback, PVOID usrPtr)
{
    BYTE r;
    BOOL notRoot;
    const HKEY key = openKey(root, path, &notRoot);
    
    if (key == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_CANT_OPEN_KEY;
    }
    
    r = enumSubKeys0(key, callback, usrPtr);
    
    if (notRoot) {
        RegCloseKey(key);
    }
    
    return r;
}


static BYTE enumValues(HKEY root, PCWSTR path,
    ValueCallback callback, PVOID usrPtr)
{
    BYTE r;
    BOOL notRoot;
    const HKEY key = openKey(root, path, &notRoot);
    
    if (key == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_CANT_OPEN_KEY;
    }
    
    r = enumValues0(key, callback, usrPtr);
    
    if (notRoot) {
        RegCloseKey(key);
    }
    
    return r;
}


static BOOL enumSubKeysCallback(PCWSTR name, PVOID usrPtr)
{
    const JNIData* const cbdPtr = (const JNIData*) usrPtr;
    JNIEnv* const e = cbdPtr->e;
    jobject cbObj = cbdPtr->o;
    
    const jclass cbClass = (*e)->GetObjectClass(e, cbObj);
    const jmethodID cbId = (*e)->GetMethodID(e, cbClass,
        "regSubKeyCallback", "(Ljava/lang/String;)Z");
 
    const jstring sName = (*e)->NewString(e, name, lstrlenW(name));
    
    return (*e)->CallBooleanMethod(e, cbObj, cbId, sName);
}


static BOOL enumValuesCallback(DWORD type, PCWSTR name, PBYTE data,
    DWORD dataLen, PVOID usrPtr)
{
    const JNIData* const cbdPtr = (const JNIData*) usrPtr;
    JNIEnv* const e = cbdPtr->e;
    jobject cbObj = cbdPtr->o;
    
    const jclass cbClass = (*e)->GetObjectClass(e, cbObj);
    const jmethodID cbId = (*e)->GetMethodID(e, cbClass,
        "regValueCallback", "(BLjava/lang/String;[B)Z");
 
    const jstring sName = (*e)->NewString(e, name, lstrlenW(name));
    const jbyteArray aData = (*e)->NewByteArray(e, dataLen);
    
    (*e)->SetByteArrayRegion(e, aData, 0, dataLen, (const jbyte*)data);
    
    return (*e)->CallBooleanMethod(e, cbObj, cbId,
        encodeValueType(type), sName, aData);
}


JNIEXPORT jbyte JNICALL Java_com_parasoft_xtest_common_nativecode_NativeRegistry_native_1queryValue
  (JNIEnv* e, jclass c, jbyte hkId, jstring sPath, jstring sName,
   jobject valueInfo)
{
    jbyte r;
    const HKEY hk = decodeHKId(hkId);
    const jchar* path;
    const jchar* name;
    DWORD type;
    PBYTE data;
    DWORD dataLen;
    
    if (hk == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_INVALID_HKID;
    }

    path = (*e)->GetStringChars(e, sPath, NULL);
    name = (*e)->GetStringChars(e, sName, NULL);
    
    r = queryValue(hk, path, name, &type, &data, &dataLen);

    (*e)->ReleaseStringChars(e, sName, name);
    (*e)->ReleaseStringChars(e, sPath, path);
    
    if (r == com_parasoft_xtest_common_nativecode_NativeRegistry_SUCCESS) {
        jfieldID id;
        jbyteArray a;
        const jclass vic = (*e)->GetObjectClass(e, valueInfo);
        
        id = (*e)->GetFieldID(e, vic, "type", "B");
        (*e)->SetByteField(e, valueInfo, id, encodeValueType(type));
        
        a = (*e)->NewByteArray(e, dataLen);
        (*e)->SetByteArrayRegion(e, a, 0, dataLen, (const jbyte*)data);
        id = (*e)->GetFieldID(e, vic, "data", "[B");
        (*e)->SetObjectField(e, valueInfo, id, a);
                 
        free(data);
    }
    
    return r;
}


JNIEXPORT jbyte JNICALL Java_com_parasoft_xtest_common_nativecode_NativeRegistry_native_1enumSubKeys
  (JNIEnv* e, jclass c, jbyte hkId, jstring sPath, jobject cbObj)
{
    jbyte r;
    const HKEY hk = decodeHKId(hkId);
    const jchar* path;
    JNIData cbd;

    cbd.e = e;
    cbd.o = cbObj;
    
    if (hk == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_INVALID_HKID;
    }
    
    path = (*e)->GetStringChars(e, sPath, NULL);

    r = enumSubKeys(hk, path, enumSubKeysCallback, &cbd);
    
    (*e)->ReleaseStringChars(e, sPath, path);
    
    return r;
}


JNIEXPORT jbyte JNICALL Java_com_parasoft_xtest_common_nativecode_NativeRegistry_native_1enumValues
  (JNIEnv* e, jclass c, jbyte hkId, jstring sPath, jobject cbObj)
{
    jbyte r;
    const HKEY hk = decodeHKId(hkId);
    const jchar* path;
    JNIData cbd;

    cbd.e = e;
    cbd.o = cbObj;
    
    if (hk == INVALID_HANDLE_VALUE) {
        return com_parasoft_xtest_common_nativecode_NativeRegistry_ERROR_INVALID_HKID;
    }
    
    path = (*e)->GetStringChars(e, sPath, NULL);

    r = enumValues(hk, path, enumValuesCallback, &cbd);
    
    (*e)->ReleaseStringChars(e, sPath, path);
    
    return r;
}


JNIEXPORT jbyteArray JNICALL Java_com_parasoft_xtest_common_nativecode_NativeRegistry_native_1expandEnvStr
  (JNIEnv* e, jclass c, jbyteArray aData)
{
	jbyteArray r = NULL;
	jbyte* srcData;
	const jsize srcDataLen = (*e)->GetArrayLength(e, aData);
	jbyte* outData;
    jsize outDataLen;
	DWORD chrNo;
	
	if (srcDataLen == 0) {
	    return aData;
	}

    srcData = malloc(srcDataLen);	
    (*e)->GetByteArrayRegion(e, aData, 0, srcDataLen, srcData);
	chrNo = ExpandEnvironmentStringsW((PWSTR)srcData, NULL, 0);

	if (chrNo) {
    	outDataLen = chrNo * sizeof(WCHAR);
    	outData = malloc(outDataLen);
	    chrNo = ExpandEnvironmentStringsW((PWSTR)srcData, (PWSTR)outData, chrNo);
	    
        if (chrNo) {
            r = (*e)->NewByteArray(e, outDataLen);
            
            if (r) {
	            (*e)->SetByteArrayRegion(e, r, 0, outDataLen, outData);
	        }
	    }
	    
	    free(outData);
    }

    free(srcData);

    return r;
}


#endif /* EXP_WIN32 */
